// VisCorBufList.cpp


#include "VisCorona.h"


CVisCoronaBuf::CVisCoronaBuf(void)
  : m_fInUse(false),
	m_milidBuf(M_NULL),
	m_evisimencoding(evisimencodingUnknown),
	m_cbData(0),
	m_pvData(0),
	m_pCorona(0)
{
	m_size.cx = m_size.cy = 0;
}


CVisCoronaBuf::~CVisCoronaBuf(void)
{
	Deallocate();
}


bool CVisCoronaBuf::FIsValid(void) const
{
	if ((m_milidBuf != M_NULL)
			&& (m_evisimencoding != evisimencodingUnknown)
			&& (m_size.cx != 0)
			&& (m_size.cy != 0)
			&& (m_cbData != 0)
			&& (m_pvData != 0)
			&& (m_pCorona != 0))
	{
		return true;
	}

	return false;
}


void CVisCoronaBuf::Allocate(CVisCorona *pCorona)
{
	assert(pCorona != 0);
	assert(pCorona->ImEncoding() != evisimencodingUnknown);

	Deallocate();

	// Note:  evisimencodingRGB555 is a different format than the Matrox
	// M_RGB15_PACKED format.
	// UNDONE:  We seem to get access violations when we try to use
	// planar YUV formats.  Maybe we're doing something wrong in this
	// class?
	// VASI: Added M_DISP to all the Buffer Attributes. Otherwise,
	// a valid host address is not returned from the call to
	// MdigInquire(m_milidBuf, M_HOST_ADDRESS, M_NULL) below.
	long cBand = 3;
	long cbitPixel;
	long lBufAttribute;
	switch (pCorona->ImEncoding())
	{
	case evisimencodingUYVY:
		lBufAttribute = M_IMAGE + M_GRAB + M_YUV16_PACKED + M_PROC + M_DISP;
		cbitPixel = 16;
		break;

#ifdef LATER
	case :
		// UNDONE:  What 16-bit planar format is being used?
		lBufAttribute = M_IMAGE + M_GRAB + M_YUV16_PLANAR + M_PROC + M_DISP;
		cbitPixel = 16;
		break;
#endif // LATER

	case evisimencodingIYUV:
		lBufAttribute = M_IMAGE + M_GRAB + M_YUV12_PLANAR + M_PROC + M_DISP;
		cbitPixel = 12;
		break;

	case evisimencodingYVU9:
		lBufAttribute = M_IMAGE + M_GRAB + M_YUV9_PLANAR + M_PROC + M_DISP;
		cbitPixel = 9;
		break;

	case evisimencodingGray8bit:
		cBand = 1;
		lBufAttribute = M_IMAGE + M_GRAB + M_PROC + M_DISP;
		cbitPixel = 8;
		break;

	default:
		// evisimencodingRGBA32
		assert(pCorona->ImEncoding() == evisimencodingRGBA32);
		lBufAttribute = M_IMAGE + M_GRAB + M_PROC + M_DISP;
		cbitPixel = 32;
		break;
	}

	// Note:  We alloc on m_milidSys for the corona because this seems to fail when
	// we try to alloc on the host.  We assume that this function gives us a buffer
	// that can be mapped to the host address space.
	m_milidBuf = MbufAllocColor(pCorona->MilidSys(), cBand, pCorona->SizeCur().cx,
			pCorona->SizeCur().cy, M_UNSIGNED + 8, lBufAttribute, M_NULL);

	// Try to alloc the buffer again if MbufAllocColor fails.
	for (int iRetry = 0; ((m_milidBuf == M_NULL) && (iRetry < 400)); ++iRetry)
	{
		Sleep(5);
		m_milidBuf = MbufAllocColor(pCorona->MilidSys(), cBand, pCorona->SizeCur().cx,
				pCorona->SizeCur().cy, M_UNSIGNED + 8, lBufAttribute, M_NULL);
	}

	if (m_milidBuf != M_NULL)
	{
		assert(!m_fInUse);

		m_pCorona = pCorona;
		m_evisimencoding = pCorona->ImEncoding();
		m_size = pCorona->SizeCur();
		m_cbData = cbitPixel * m_size.cx * m_size.cy;
		m_pvData = (void *) MbufInquire(m_milidBuf, M_HOST_ADDRESS, M_NULL);

#ifdef _DEBUG
		IncCBufInUse();
	}
	else
	{
		static bool fAsserted = false;
		assert(fAsserted);
		fAsserted = true;
#endif // _DEBUG
	}
}

void CVisCoronaBuf::Deallocate(void)
{
	assert(!m_fInUse);

	assert(M_NULL == 0);
	if (m_milidBuf != M_NULL)
	{
		MbufFree(m_milidBuf);

		m_milidBuf = M_NULL;
		m_pCorona = 0;
		m_evisimencoding = evisimencodingUnknown;
		m_size.cx = m_size.cy = 0;
		m_cbData = 0;
		m_pvData = 0;

#ifdef _DEBUG
		DecCBufInUse();
	}
	else
	{
		assert(m_milidBuf == M_NULL);
		assert(m_evisimencoding == evisimencodingUnknown);
		assert(m_size.cx == 0);
		assert(m_size.cy == 0);
		assert(m_cbData == 0);
		assert(m_pvData == 0);
		assert(m_pCorona == 0);
#endif // _DEBUG
	}
}


// UNDONE:  Use crit sect in buf list


int CVisCoronaBufList::CBufCur(const CVisCorona *pCorona)
{
	int cbuf = 0;

	EnterCriticalSection(&m_criticalsection);

	if (!m_vectorpCoronaBuf.empty())
	{
		std::vector<CVisCoronaBuf *>::const_iterator iteratorCur = m_vectorpCoronaBuf.begin();
		std::vector<CVisCoronaBuf *>::const_iterator iteratorLim = m_vectorpCoronaBuf.end();
		for ( ; iteratorCur != iteratorLim; ++iteratorCur)
		{
			if ((pCorona == 0)
					|| (pCorona == (*iteratorCur)->PCorona()))
			{
				++ cbuf;
			}
		}
	}

	LeaveCriticalSection(&m_criticalsection);

	return cbuf;
}

int CVisCoronaBufList::CBufInUse(const CVisCorona *pCorona)
{
	int cbuf = 0;

	EnterCriticalSection(&m_criticalsection);

	if (!m_vectorpCoronaBuf.empty())
	{
		std::vector<CVisCoronaBuf *>::const_iterator iteratorCur = m_vectorpCoronaBuf.begin();
		std::vector<CVisCoronaBuf *>::const_iterator iteratorLim = m_vectorpCoronaBuf.end();
		for ( ; iteratorCur != iteratorLim; ++iteratorCur)
		{
			if (((pCorona == 0)
						|| (pCorona == (*iteratorCur)->PCorona()))
					&& ((*iteratorCur)->FIsInUse()))
			{
				++ cbuf;
			}
		}
	}

	LeaveCriticalSection(&m_criticalsection);

	return cbuf;
}

int CVisCoronaBufList::CBufUnused(const CVisCorona *pCorona)
{
	int cbuf = 0;

	EnterCriticalSection(&m_criticalsection);

	if (!m_vectorpCoronaBuf.empty())
	{
		std::vector<CVisCoronaBuf *>::const_iterator iteratorCur = m_vectorpCoronaBuf.begin();
		std::vector<CVisCoronaBuf *>::const_iterator iteratorLim = m_vectorpCoronaBuf.end();
		for ( ; iteratorCur != iteratorLim; ++iteratorCur)
		{
			if (((pCorona == 0)
						|| (pCorona == (*iteratorCur)->PCorona()))
					&& (!(*iteratorCur)->FIsInUse()))
			{
				++ cbuf;
			}
		}
	}

	LeaveCriticalSection(&m_criticalsection);

	return cbuf;
}


CVisCoronaBuf *CVisCoronaBufList::PBufGetPCorona(CVisCorona *pCorona)
{
	assert(pCorona != 0);
	
	CVisCoronaBuf *pBuf = 0;

	EnterCriticalSection(&m_criticalsection);

	if (!m_vectorpCoronaBuf.empty())
	{
		std::vector<CVisCoronaBuf *>::const_iterator iteratorCur = m_vectorpCoronaBuf.begin();
		std::vector<CVisCoronaBuf *>::const_iterator iteratorLim = m_vectorpCoronaBuf.end();
		for ( ; iteratorCur != iteratorLim; ++iteratorCur)
		{
			if ((pCorona == (*iteratorCur)->PCorona())
					&& (!(*iteratorCur)->FIsInUse()))
			{
				pBuf = (*iteratorCur);
				break;
			}
		}
	}

	if (pBuf != 0)
	{
		pBuf->SetFIsInUse(true);
		pBuf->PCorona()->AddRef();
	}

	LeaveCriticalSection(&m_criticalsection);

	if (pBuf == 0)
	{
		pBuf = new CVisCoronaBuf;
		if (pBuf != 0)  // LATER:  Throw if OOM?
		{
			pBuf->Allocate(pCorona);
			if (pBuf->FIsValid())
			{
				EnterCriticalSection(&m_criticalsection);

				m_vectorpCoronaBuf.push_back(pBuf);
				pBuf->SetFIsInUse(true);
				pBuf->PCorona()->AddRef();

				LeaveCriticalSection(&m_criticalsection);
			}
			else
			{
				// Fail.
				delete pBuf;
				pBuf = 0;
			}
		}
	}

	return pBuf;
}


void CVisCoronaBufList::ReleaseBuf(CVisCoronaBuf *pBuf)
{
	assert(pBuf != 0);

	EnterCriticalSection(&m_criticalsection);

	assert(pBuf->FIsInUse());

	pBuf->SetFIsInUse(false);

	CVisCorona *pCorona = pBuf->PCorona();
	if ((pBuf->ImEncoding() != pCorona->ImEncoding())
			|| (pBuf->Size().cx != pCorona->SizeCur().cx)
			|| (pBuf->Size().cy != pCorona->SizeCur().cy))
	{
		RemoveBuf(pBuf);
	}

	pCorona->Release();

	LeaveCriticalSection(&m_criticalsection);
}


void VISAPI CVisCoronaBufList::StaticReleaseBuf(void *pvData, void *pvBuf)
{
	assert(pvData != 0);
	assert(pvBuf != 0);

	CVisCoronaProvider *pCoronaProvider
			= CVisCoronaProvider::PCoronaProviderGet();

	assert(pCoronaProvider != 0);

	pCoronaProvider->CoronaBufList().ReleaseBuf((CVisCoronaBuf *) pvBuf);

	pCoronaProvider->Release();
}


void CVisCoronaBufList::ReallocBufs(CVisCorona *pCorona, int cBuf)
{
	assert(pCorona != 0);

	EnterCriticalSection(&m_criticalsection);

	int cBufValid = 0;
	if (!m_vectorpCoronaBuf.empty())
	{
		// Note:  We'd like to use the reverse iterator to avoid extra copies, but the
		// erase() method takes an interator, not a reverse_iterator.
		std::vector<CVisCoronaBuf *>::iterator iteratorCur = m_vectorpCoronaBuf.begin();
		std::vector<CVisCoronaBuf *>::iterator iteratorLim = m_vectorpCoronaBuf.end();
		
		for ( ; iteratorCur != iteratorLim; ++iteratorCur)
		{
			CVisCoronaBuf *pBuf = (*iteratorCur);
			if ((pCorona == pBuf->PCorona()) && (!pBuf->FIsInUse()))
			{
				if ((pBuf->ImEncoding() != pCorona->ImEncoding())
						|| (pBuf->Size().cx != pCorona->SizeCur().cx)
						|| (pBuf->Size().cy != pCorona->SizeCur().cy))
				{
					// Destroy the buffer.
					delete pBuf;
					
					// Remove the pointer to the buffer from the list.
					m_vectorpCoronaBuf.erase(iteratorCur);

					// Adjust the pointers used in the loop.
					-- iteratorCur;
					-- iteratorLim;
				}
				else if (++cBufValid > cBuf)
				{
//					-- cBuf;		// VASI: Shouldn't this be -- cBufValid??
					-- cBufValid;	// VASI: UNDONE

					// Destroy the buffer.
					delete pBuf;
					
					// Remove the pointer to the buffer from the list.
					m_vectorpCoronaBuf.erase(iteratorCur);

					// Adjust the pointers used in the loop.
					-- iteratorCur;
					-- iteratorLim;
				}
			}
		}
	}

	LeaveCriticalSection(&m_criticalsection);

	// Try to allocate more buffers, if needed.
	while (cBufValid < cBuf)
	{
		CVisCoronaBuf *pBuf = new CVisCoronaBuf;
		if (pBuf == 0)
			break;  // LATER:  Throw if OOM?

		pBuf->Allocate(pCorona);
		if (pBuf->FIsValid())
		{
			EnterCriticalSection(&m_criticalsection);

			m_vectorpCoronaBuf.push_back(pBuf);
			++ cBufValid;

			LeaveCriticalSection(&m_criticalsection);
		}
		else
		{
			// Fail.
			delete pBuf;
			break;
		}
	}
}

void CVisCoronaBufList::RemoveBufs(const CVisCorona *pCorona)
{
	EnterCriticalSection(&m_criticalsection);

	if (!m_vectorpCoronaBuf.empty())
	{
		// Note:  We'd like to use the reverse iterator to avoid extra copies, but the
		// erase() method takes an interator, not a reverse_iterator.
		std::vector<CVisCoronaBuf *>::iterator iteratorCur = m_vectorpCoronaBuf.begin();
		std::vector<CVisCoronaBuf *>::iterator iteratorLim = m_vectorpCoronaBuf.end();
		
		for ( ; iteratorCur != iteratorLim; ++iteratorCur)
		{
			if (((pCorona == 0)
						|| (pCorona == (*iteratorCur)->PCorona()))
					&& (!(*iteratorCur)->FIsInUse()))
			{
				// Destroy the buffer.
				delete (*iteratorCur);
				
				// Remove the pointer to the buffer from the list.
				m_vectorpCoronaBuf.erase(iteratorCur);

				// Adjust the pointers used in the loop.
				-- iteratorCur;
				-- iteratorLim;
			}
		}
	}

	LeaveCriticalSection(&m_criticalsection);
}

void CVisCoronaBufList::RemoveBuf(CVisCoronaBuf *pCoronaBuf)
{
	assert(pCoronaBuf != 0);

	EnterCriticalSection(&m_criticalsection);

	assert(!(pCoronaBuf->FIsInUse()));

	if (!m_vectorpCoronaBuf.empty())
	{
		std::vector<CVisCoronaBuf *>::iterator iteratorCur = m_vectorpCoronaBuf.begin();
		std::vector<CVisCoronaBuf *>::iterator iteratorLim = m_vectorpCoronaBuf.end();
		for ( ; iteratorCur != iteratorLim; ++iteratorCur)
		{
			if (pCoronaBuf == (*iteratorCur))
			{
				// Remove the pointer to the buffer from the list.
				m_vectorpCoronaBuf.erase(iteratorCur);
				// We don't adjust the loop pointers here because we're
				// going to break out of the loop.
				break;
			}
		}
	}

	delete pCoronaBuf;

	LeaveCriticalSection(&m_criticalsection);
}



#ifdef _DEBUG

int CVisCoronaBuf::s_cBufInUseCur = 0;
int CVisCoronaBuf::s_cBufInUseHigh = 0;


// UNDONE:  Add an assert that s_cBufInUseCur is 0 on exit.


void CVisCoronaBuf::IncCBufInUse(void)
{
	if (++s_cBufInUseCur > s_cBufInUseHigh)
	{
		assert(s_cBufInUseCur == s_cBufInUseHigh + 1);
		++s_cBufInUseHigh;
	}
}


void CVisCoronaBuf::DecCBufInUse(void)
{
	--s_cBufInUseCur;
	assert(s_cBufInUseCur >= 0);
}
#endif // _DEBUG
